package no.ntnu.ime.xsltgenerator.client.ui;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map.Entry;
import no.ntnu.ime.xsltgenerator.client.models.Relation;
import no.ntnu.ime.xsltgenerator.client.models.RelationType;
import no.ntnu.ime.xsltgenerator.client.models.TreeNodeValidationItem;
import no.ntnu.ime.xsltgenerator.client.services.XsltUtils;

import com.google.gwt.core.client.GWT;
import com.google.gwt.event.dom.client.ChangeEvent;
import com.google.gwt.event.dom.client.ChangeHandler;
import com.google.gwt.event.dom.client.ClickEvent;
import com.google.gwt.event.dom.client.ClickHandler;
import com.google.gwt.event.logical.shared.SelectionEvent;
import com.google.gwt.event.logical.shared.SelectionHandler;
import com.google.gwt.safehtml.shared.SafeHtmlUtils;
import com.google.gwt.user.client.Window;
import com.google.gwt.user.client.ui.Button;
import com.google.gwt.user.client.ui.Composite;
import com.google.gwt.user.client.ui.FlexTable;
import com.google.gwt.user.client.ui.HTML;
import com.google.gwt.user.client.ui.HorizontalPanel;
import com.google.gwt.user.client.ui.Label;
import com.google.gwt.user.client.ui.ListBox;
import com.google.gwt.user.client.ui.Tree;
import com.google.gwt.user.client.ui.TreeItem;
import com.google.gwt.user.client.ui.VerticalPanel;
import com.google.gwt.xml.client.NamedNodeMap;
import com.google.gwt.xml.client.Node;
import com.google.gwt.xml.client.NodeList;

/***
 * User interface for manually mappping xml-elements to their meaning in an
 * OpenFEIDE context.
 * 
 * @author Kristian Klette <kriskl@stud.ntnu.no>
 * 
 *         <h2>Relation management</h2> The mapper support three (3) kinds of
 *         relationship mappings:
 *         <ul>
 *         <li>organization &lt;-- group</li>
 *         <li>organization &lt;-- person</li>
 *         <li>group &lt;-- person</li>
 *         </li> In addition to pure affiliations, the mappings might add
 *         additional affiliations to the FEIDE predefined roles:
 *         <ul>
 *         <li>MEMBER
 *         <ul>
 *         <li>STUDENT</li>
 *         <li>EMPLOYEE
 *         <ul>
 *         <li>FACULTY</li>
 *         <li>STAFF</li>
 *         </ul>
 *         </li>
 *         </ul>
 *         </li>
 *         <li>AFFILIATE</li>
 *         <li>ALUM</li>
 *         <li>LIBRARY-WALK-IN</li>
 *         </ul>
 * 
 *         These roles are added if the user selects the appropriate fields in
 *         the "Ekstra:"-list when defining the membership element. Each role
 *         implies that the user has all the affiliations further up the tree.
 *         Ie. a member of the "STAFF"-role are implicitly added the roles of
 *         "EMPLOYEE" and "MEMBER".
 * 
 */
public class ManualConversionPanel extends Composite {

	private static final String ORGANIZATION_UNIT_TYPE_IDENTIFIER = "organizationUnit";
	private static final String ORGANIZATION_TYPE_IDENTIFIER = "organization";
	private static final String MEMBERSHIP_TYPE_IDENTIFIER = "membership";
	private static final String GROUP_TYPE_IDENTIFIER = "group";
	private static final String PERSON_TYPE_IDENTIFIER = "person";

	private VerticalPanel verticalPanel;

	private final ArrayList<HashMap<String, String>> persons = new ArrayList<HashMap<String, String>>();
	private final ArrayList<HashMap<String, String>> groups = new ArrayList<HashMap<String, String>>();

	private String personNodePath = "";
	private String groupsNodePath = "";

	private final HashMap<String, TreeItem> xmlPathCache = new HashMap<String, TreeItem>();
	private final HashMap<String, String[]> typeToElementMap = new HashMap<String, String[]>();
	private HashMap<String, String> pathToTypeMap = new HashMap<String, String>();

	private ListBox elementTypesList;
	private Label selectedElementTypeLabel;
	private Label selectedElementAttributes;
	private Label selectedElementValue;
	private Label selectedElementName;
	private ListBox elementSpecificTypesList;

	public TreeNodeValidationItem currentSelectedTreeNodeValidationItem;

	private Tree xmlTreeRepresentation;

	private HTML xlstPreview;
	private ListBox additionalAffiliations;
	public ArrayList<Relation> relations = new ArrayList<Relation>();

	public ManualConversionPanel(Node sourceXML) {
		if (sourceXML == null) {
			throw new IllegalArgumentException("sourceXML cannot be null");
		}
		populateTypeToElementMap();

		verticalPanel = new VerticalPanel();
		initWidget(verticalPanel);
		verticalPanel.setSize("100%", "100%");

		Label label = new Label(
				"Velg først en verdi i tre-strukturen på venstre side, "
						+ "og en tilhørende type i listen på høyre side. Trykk så på \"Lagre valg\"."
						+ "Nederst på siden vil det bli generert en forhåndsvisning av XSLT-transformasjonen.");
		verticalPanel.add(label);

		HorizontalPanel horizontalPanel = new HorizontalPanel();
		verticalPanel.add(horizontalPanel);
		horizontalPanel.setSize("100%", "100%");

		xmlTreeRepresentation = new Tree();
		horizontalPanel.add(xmlTreeRepresentation);
		xmlTreeRepresentation.setWidth("100%");

		xmlTreeRepresentation
				.addSelectionHandler(new SelectionHandler<TreeItem>() {
					@Override
					public void onSelection(final SelectionEvent<TreeItem> event) {
						currentSelectedTreeNodeValidationItem = (TreeNodeValidationItem) event
								.getSelectedItem().getUserObject();
						selectedElementName
								.setText(currentSelectedTreeNodeValidationItem
										.getText());
						selectedElementValue
								.setText(currentSelectedTreeNodeValidationItem
										.getValue());
						selectedElementAttributes.setText("");
						selectedElementAttributes
								.setText(currentSelectedTreeNodeValidationItem
										.getAttributesAsString());
					}
				});

		TreeItem rootItem = new TreeItem();
		xmlTreeRepresentation.addItem(rootItem);
		buildTreeRepresentation(sourceXML.getFirstChild(), rootItem,
				sourceXML.getNodeName());

		FlexTable flexTable = new FlexTable();
		horizontalPanel.add(flexTable);
		flexTable.setWidth("100%");

		Label selectedElementNameLabel = new Label("Valgt element:");
		flexTable.setWidget(0, 0, selectedElementNameLabel);

		selectedElementName = new Label("");
		flexTable.setWidget(0, 1, selectedElementName);

		Label selectedElementValueLabel = new Label("Verdi:");
		flexTable.setWidget(1, 0, selectedElementValueLabel);

		selectedElementValue = new Label("");
		flexTable.setWidget(1, 1, selectedElementValue);

		Label selectedElementAttributesLabel = new Label("Attributter:");
		flexTable.setWidget(2, 0, selectedElementAttributesLabel);

		selectedElementAttributes = new Label("");
		flexTable.setWidget(2, 1, selectedElementAttributes);

		selectedElementTypeLabel = new Label("Velg type:");
		flexTable.setWidget(3, 0, selectedElementTypeLabel);

		elementTypesList = new ListBox();
		elementTypesList.setVisibleItemCount(1);
		elementTypesList.addItem("");
		elementTypesList.addItem("Person", PERSON_TYPE_IDENTIFIER);
		elementTypesList.addItem("Gruppering", GROUP_TYPE_IDENTIFIER);
		elementTypesList.addItem("Medlemskap", MEMBERSHIP_TYPE_IDENTIFIER);
		elementTypesList.addItem("Organisasjon", ORGANIZATION_TYPE_IDENTIFIER);
		elementTypesList.addItem("Org. enhet",
				ORGANIZATION_UNIT_TYPE_IDENTIFIER);
		elementTypesList.addChangeHandler(new ElementTypeListChangeHandler());
		flexTable.setWidget(3, 1, elementTypesList);

		Label elementSpecificTypeLabel = new Label("Beskrivelse av type:");
		flexTable.setWidget(4, 0, elementSpecificTypeLabel);

		elementSpecificTypesList = new ListBox();
		elementSpecificTypesList.setVisibleItemCount(1);
		flexTable.setWidget(4, 1, elementSpecificTypesList);

		Label additionalAffiliationsLabel = new Label("Ekstra:");
		flexTable.setWidget(5, 0, additionalAffiliationsLabel);

		additionalAffiliations = new ListBox(true);
		flexTable.setWidget(5, 1, additionalAffiliations);
		additionalAffiliations.addItem("Medlemskap angir rolle som student",
				"add_student");
		additionalAffiliations
				.addItem(
						"Medlemskap angir rolle som vitenskapelig ansatt (lærer, professor osv.)",
						"add_faculty");
		additionalAffiliations.addItem(
				"Medlemskap angir rolle som ikke vitenskapelig ansatt",
				"add_staff");
		additionalAffiliations.addItem("Medlemskap angir rolle som ekstern",
				"add_affiliated");
		additionalAffiliations.addItem(
				"Medlemskap angir rolle som \"tilgang til bibliotek\"",
				"add_library_walk_in");

		// Is enabled upon selection of membership type
		additionalAffiliations.setEnabled(false);

		Button button = new Button("Lagre valg");
		flexTable.setWidget(6, 0, button);
		button.setWidth("100%");
		button.addClickHandler(new SaveSelectionButtonClickHandler());
		flexTable.getFlexCellFormatter().setColSpan(6, 0, 2);

		xlstPreview = new HTML();
		verticalPanel.add(xlstPreview);
	}

	private void addGroupTypeToList(ListBox listBox) {
		listBox.addItem("Gruppering", "grp_group_group");
		listBox.addItem("Gruppering > Gruppe ID", "val_group_id");
		listBox.addItem("Gruppering > Beskrivelse", "val_group_description");
	}

	private void addPersonTypesToList(ListBox listBox) {
		listBox.addItem("Person (node med barn)", "grp_person_group");
		listBox.addItem("PersonID", "val_person_id");
		listBox.addItem("Fornavn", "val_person_fornavn");
		listBox.addItem("Etternavn", "val_person_etternavn");
		listBox.addItem("Fullt navn", "val_person_fullt_navn");
		listBox.addItem("Fødselsdato", "val_person_fodt");
		listBox.addItem("Epost", "val_person_epost");
		listBox.addItem("Mobil", "val_person_mobil");
		listBox.addItem("Fasttelefon", "val_person_tlf");
		listBox.addItem("Brukernavn", "val_person_brukernavn");
		listBox.addItem("Adresse (node med barn)", "grp_person_adresse");
		listBox.addItem("Gateadresse", "val_person_gateadresse");
		listBox.addItem("Postnummer", "val_person_postnr");
		listBox.addItem("By", "val_person_by");
		listBox.addItem("Land", "val_person_land");
	}

	private void addRelationTypesToList(ListBox listBox) {
		listBox.addItem("Medlemskap (node med barn)", "grp_relation");

		listBox.addItem("Mål", "grp_relation_target");

		listBox.addItem("Mål > Organisasjon", "grp_relation_target_org");
		listBox.addItem("Mål > Organisasjon > Organisasjons ID",
				"val_relation_target_org_orgid");
		listBox.addItem("Mål > Organisasjon > Enhet",
				"val_relation_target_org_ou");

		listBox.addItem("Mål > Gruppering", "grp_relation_target_group");
		listBox.addItem("Mål > Gruppering > Gruppe ID",
				"val_relation_target_group_id");

		listBox.addItem("Medlem", "grp_relation_object");
		listBox.addItem("Medlem > Person ID", "val_relation_object_personid");
	}

	private void buildTreeRepresentation(final Node node,
			final TreeItem parentTreeNode, final String currentPath) {
		if (node == null) {
			return;
		}
		// Fast-path exit for empty text-nodes
		if ((node.getNodeType() == Node.TEXT_NODE)
				&& node.getNodeValue().trim().isEmpty()) {
			return;
		}

		if (node.getNodeType() != Node.TEXT_NODE) {
			// GWT.log("BTP: + " + node.getNodeName());
			try {
				final HashMap<String, String> itemAttributes = getAttributeMapForNode(node);
				final TreeNodeValidationItem item = new TreeNodeValidationItem(
						node.getNodeName(), itemAttributes, false);
				TreeItem treeNode = new TreeItem(item);
				treeNode.setUserObject(item);
				treeNode.setState(true);

				// See if we need to add this node. If the same node is
				// represented
				// in the tree, then we just use
				// it as the parent for the recursion.
				// We have to add the node to the tree for the duplicate check -
				// even if it's just to remove it again.

				parentTreeNode.addItem(treeNode);
				final String path = get_tree_path(treeNode);
				if (xmlPathCache.containsKey(path)) {
					// Duplicate - remove the node
					parentTreeNode.removeItem(treeNode);
					treeNode = xmlPathCache.get(path);

				} else {
					xmlPathCache.put(path, treeNode);
				}

				// Recurse further down the tree
				final NodeList children = node.getChildNodes();
				for (int childIndex = 0; childIndex < children.getLength(); childIndex++) {
					final Node child = children.item(childIndex);
					buildTreeRepresentation(child, treeNode, path);
				}
			} catch (Exception e) {
				GWT.log("Error while building tree", e);
			}

		} else {
			// Add pure text-nodes to the parent's userObject data value
			final TreeNodeValidationItem parentItem = (TreeNodeValidationItem) parentTreeNode
					.getUserObject();
			parentItem.setDataValue(node.getNodeValue());

			// Update the UI by updating the parents HTML-content.
			parentTreeNode.setHTML(parentItem);
		}
	}

	private String generateXLST() {
		final StringBuilder xlst = new StringBuilder();

		if (!personNodePath.isEmpty()) {
			xlst.append(XsltUtils.xsltFragmentForStandardMapping(
					personNodePath, "person", persons, typeToElementMap));
		}
		if (!groupsNodePath.isEmpty()) {
			xlst.append(XsltUtils.xsltFragmentForStandardMapping(
					groupsNodePath, "OrganizationUnit", groups,
					typeToElementMap));
		}

		if (!relations.isEmpty()) {
			for (Relation r : relations) {
				xlst.append(r.personAffiliationXslt());
			}
		}

		return xlst.toString();
	}

	private String get_tree_path(final TreeItem selectedItem) {
		final StringBuilder pathBuilder = new StringBuilder();
		final TreeNodeValidationItem item = (TreeNodeValidationItem) selectedItem
				.getUserObject();
		if (item == null) {
			return "/";
		}
		if (!item.getAttributes().keySet().isEmpty()) {
			final String attrString = getAttributeXPathString(item);

			pathBuilder.append(item.getText() + attrString + "/");
		} else {
			pathBuilder.append(item.getText() + "/");
		}
		TreeItem parent = selectedItem.getParentItem();

		while (parent != null) {
			if (parent.getUserObject() != null) {
				final TreeNodeValidationItem pItem = (TreeNodeValidationItem) parent
						.getUserObject();
				if (!pItem.getAttributes().keySet().isEmpty()) {
					final String attrString = getAttributeXPathString(pItem);
					pathBuilder.append(pItem.getText() + attrString + "/");
				} else {
					pathBuilder.append(pItem.getText() + "/");
				}
			} else {
				pathBuilder.append("/");
			}
			parent = parent.getParentItem();
		}

		final String path = pathBuilder.toString();

		final String[] parts = path.split("/");
		String reversed = "";
		for (final String part : parts) {
			reversed = part + "/" + reversed;
		}
		if (!reversed.startsWith("/")) {
			reversed = "/" + reversed;
		}
		if (reversed.endsWith("/")) {
			reversed = reversed.substring(0, reversed.length() - 1);
		}
		return reversed;
	}

	private String getAttributeXPathString(final TreeNodeValidationItem item) {
		final StringBuilder attrStringBuilder = new StringBuilder();
		attrStringBuilder.append("[");
		int attrsRemaining = item.getAttributes().size();
		for (final Entry<String, String> attr : item.getAttributes().entrySet()) {
			if (attrsRemaining > 1) {
				attrStringBuilder.append("@" + attr.getKey() + "='"
						+ attr.getValue() + "' and");
			} else {
				attrStringBuilder.append("@" + attr.getKey() + "='"
						+ attr.getValue() + "'");
			}
			attrsRemaining--;
		}
		attrStringBuilder.append("]");
		final String attrString = attrStringBuilder.toString();
		return attrString;
	}

	private HashMap<String, String> getAttributeMapForNode(final Node child) {
		final HashMap<String, String> itemAttributes = new HashMap<String, String>();

		if (child.getNodeType() == Node.ELEMENT_NODE) {
			final NamedNodeMap attrs = child.getAttributes();
			for (int i = 0; i < attrs.getLength(); i++) {
				final Node attr = attrs.item(i);
				if ((attr.getNodeName() != null)
						&& (attr.getNodeValue() != null)) {
					itemAttributes.put(attr.getNodeName(), attr.getNodeValue());
				}
			}
		}
		return itemAttributes;
	}

	private void populateTypeToElementMap() {

		// person elements
		typeToElementMap.put("val_person_id", new String[] {
				"of:norEduPersonNIN", "of:uid" });
		typeToElementMap.put("val_person_fornavn",
				new String[] { "of:givenName" });
		typeToElementMap.put("val_person_fullt_navn",
				new String[] { "of:norEduPersonLegalName" });
		typeToElementMap.put("val_person_etternavn", new String[] { "of:sn" });
		typeToElementMap.put("val_person_epost", new String[] { "of:email" });
		typeToElementMap.put("val_person_mobil", new String[] { "of:mobile" });
		typeToElementMap.put("val_person_telefon",
				new String[] { "of:homePhone" });

		// groups are treated as OUs
		typeToElementMap.put("val_group_id",
				new String[] { "of:norEduOrgUnitUniqueIdentifier" });
		typeToElementMap.put("val_group_description", new String[] { "of:ou" });

	}

	class ElementTypeListChangeHandler implements ChangeHandler {

		@Override
		public void onChange(ChangeEvent event) {
			elementSpecificTypesList.clear();
			additionalAffiliations.setEnabled(false);
			int selectedElement = elementTypesList.getSelectedIndex();
			if (selectedElement == -1) {
				return;
			} else {
				String selectedType = elementTypesList
						.getValue(selectedElement);
				if (selectedType.equals(PERSON_TYPE_IDENTIFIER)) {
					addPersonTypesToList(elementSpecificTypesList);
				} else if (selectedType.equals(GROUP_TYPE_IDENTIFIER)) {
					addGroupTypeToList(elementSpecificTypesList);
				} else if (selectedType.equals(MEMBERSHIP_TYPE_IDENTIFIER)) {
					additionalAffiliations.setEnabled(true);
					addRelationTypesToList(elementSpecificTypesList);
				} else if (selectedType.equals(ORGANIZATION_TYPE_IDENTIFIER)) {

				} else if (selectedType
						.equals(ORGANIZATION_UNIT_TYPE_IDENTIFIER)) {

				}
			}

		}
	}

	class SaveSelectionButtonClickHandler implements ClickHandler {

		@Override
		public void onClick(final ClickEvent event) {

			if (currentSelectedTreeNodeValidationItem != null) {

				int selectedIndex = elementSpecificTypesList.getSelectedIndex();
				String selectedType = elementSpecificTypesList
						.getValue(selectedIndex);

				if (xmlTreeRepresentation.getSelectedItem().getChildCount() > 0) {
					if (selectedType.startsWith("val_")) {
						Window.alert(selectedType
								+ " kan kun benyttes p� elementer som ikke har underelementer.");
						return;
					}
				} else {
					if (selectedType.startsWith("grp_")) {
						Window.alert(selectedType
								+ " kan kun benyttes p� elementer som har underelementer.");
						return;
					}
				}
				// Update the check box to be marked as done
				currentSelectedTreeNodeValidationItem.setStatus(true);
				xmlTreeRepresentation.getSelectedItem().setHTML(
						currentSelectedTreeNodeValidationItem);

				final String path = get_tree_path(xmlTreeRepresentation
						.getSelectedItem());
				pathToTypeMap.put(path, selectedType);

				if (selectedType.equals("grp_person_group")) {
					personNodePath = path;
				} else if (selectedType.equals("grp_group_group")) {
					groupsNodePath = path;
				} else {
					final String[] typeParts = selectedType.split("_");
					final String type = typeParts[1];
					final HashMap<String, String> entry = new HashMap<String, String>();
					entry.put(path, selectedType);
					if (type.equals("person")) {
						persons.add(entry);
					} else if (type.equals("group")) {
						groups.add(entry);

					} else if (type.equals("relation")) {

						if (selectedType.equals("grp_relation")) {
							String identifier = path;
							List<String[]> roles = new ArrayList<String[]>();

							for (int index = 0; index < additionalAffiliations
									.getItemCount(); index++) {
								if (additionalAffiliations
										.isItemSelected(index)) {

									String role = additionalAffiliations
											.getValue(index);
									List<String> roleTree = getImplisitRolesForRole(role);
									String attr = getAttributeXPathString(currentSelectedTreeNodeValidationItem);
									for (String r : roleTree) {
										roles.add(new String[] { r, attr });
									}
								}
							}
							Relation rel = new Relation(
									RelationType.PERSON_IN_GROUP, identifier,
									roles);
							relations.add(rel);
						} else {
							if (selectedType
									.equals("val_relation_target_org_orgid")) {
								// Find relation
								Relation rel = findRelationForPath(path);
								// Add target org_id
								rel.addTarget(currentSelectedTreeNodeValidationItem
										.getValue());
							}
						}

					}
				}
				xlstPreview.setHTML("<pre>"
						+ SafeHtmlUtils.htmlEscape(generateXLST()) + "</pre>");

			}
		}

		private Relation findRelationForPath(String path) {
			Relation maxRel = null;
			String maxPath = "";
			for (Relation rel : relations) {
				if (path.startsWith(rel.getIdentifier())
						&& rel.getIdentifier().length() > maxPath.length()) {
					maxPath = rel.getIdentifier();
					maxRel = rel;
				}
			}
			return maxRel;
		}

		/***
		 * Return a list of affiliations that should be added based upon a role.
		 * 
		 * @param role
		 * @return
		 */
		private List<String> getImplisitRolesForRole(String role) {

			// Hard coded, but the tree isn't that big
			if (role.equals("add_student")) {
				return Arrays.asList("student", "member");
			} else if (role.equals("add_faculty")) {
				return Arrays.asList("faculty", "employee", "member");
			} else if (role.equals("add_staff")) {
				return Arrays.asList("staff", "employee", "member");
			} else if (role.equals("add_affiliated")) {
				return Arrays.asList("affiliate");
			} else if (role.equals("add_library_walk_in")) {
				return Arrays.asList("library-walk-in");
			}
			ArrayList<String> roles = new ArrayList<String>();
			return roles;
		}
	}

}
