package no.ntnu.ime.libxsltgenerator;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.logging.Logger;

import no.ntnu.ime.libxsltgenerator.models.SaxParserResult;
import no.ntnu.ime.libxsltgenerator.parsers.GroupSAXParser;
import no.ntnu.ime.libxsltgenerator.parsers.OrganizationSAXParser;
import no.ntnu.ime.libxsltgenerator.parsers.PersonSAXParser;
import no.ntnu.ime.libxsltgenerator.parsers.RelationSAXParser;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.XMLReaderFactory;

public class SAXHandler extends DefaultHandler {

	private static final Logger LOGGER = Logger.getLogger(SAXHandler.class.getName());
	private final PersonSAXParser personParser = new PersonSAXParser();
	private final GroupSAXParser groupParser = new GroupSAXParser();
	private final OrganizationSAXParser organizationParser = new OrganizationSAXParser();
	private final RelationSAXParser relationParser = new RelationSAXParser();
	
	private long startElementTime = 0;
	private long endElementTime = 0;
	private long charTime = 0;
	private long endDocumentTime = 0;
	private long startDocumentTime = 0;
	
	@Override
	public void startElement(String uri, String localName, String qName,
			Attributes attributes) throws SAXException {
		long t = System.nanoTime();
		personParser.parseStartElement(uri, localName, qName, attributes);
		groupParser.parseStartElement(uri, localName, qName, attributes);
		organizationParser.parseStartElement(uri, localName, qName, attributes);
		relationParser.parseStartElement(uri, localName, qName, attributes);
		startElementTime += System.nanoTime() - t;
		
	}

	@Override
	public void endElement(String uri, String localName, String qName)
			throws SAXException {
		long t = System.nanoTime();
		personParser.parseEndElement(uri, localName, qName);
		groupParser.parseEndElement(uri, localName, qName);
		organizationParser.parseEndElement(uri, localName, qName);
		relationParser.parseEndElement(uri, localName, qName);
		endElementTime += System.nanoTime() - t;
	}

	@Override
	public void characters(char[] ch, int start, int length)
			throws SAXException {
		long t = System.nanoTime();
		String content = new String(ch, start, length);
		if (content.trim().isEmpty())
			return;

		personParser.parseCharacterData(content);
		groupParser.parseCharacterData(content);
		organizationParser.parseCharacterData(content);
		relationParser.parseCharacterData(content);
		charTime += System.nanoTime() - t;
	}

	@Override
	public void startDocument() throws SAXException {
		long t = System.nanoTime();
		personParser.startDocumentParsing();
		groupParser.startDocumentParsing();
		organizationParser.startDocumentParsing();
		relationParser.startDocumentParsing();
		startDocumentTime += System.nanoTime() - t;
	};

	@Override
	public void endDocument() throws SAXException {

		long t = System.nanoTime();
		personParser.endDocumentParsing();
		groupParser.endDocumentParsing();
		organizationParser.endDocumentParsing();
		relationParser.endDocumentParsing();

		endDocumentTime += System.nanoTime() - t;
	}

	public Map<String,List<SaxParserResult>> processSource(File file) throws Exception {
		HashMap<String, List<SaxParserResult>> allResults = new HashMap<String, List<SaxParserResult>>();
		
		// Return empty result set if any error occurs. 
		try {
			
			XMLReader parser = XMLReaderFactory.createXMLReader();
			parser.setContentHandler(this);
			parser.parse(new InputSource(new InputStreamReader(new FileInputStream(file))));
			
			long t = System.nanoTime();
			SaxParserResult personResult = personParser.getParserResult(new InputSource(new InputStreamReader(new FileInputStream(file))));
			SaxParserResult groupResult = groupParser.getParserResult(new InputSource(new InputStreamReader(new FileInputStream(file))));
			SaxParserResult organizationResult = organizationParser.getParserResult(new InputSource(new InputStreamReader(new FileInputStream(file))));
			List<SaxParserResult> relationResult = relationParser.getAllResults(file);
			long parserTime = System.nanoTime() - t;
			LOGGER.info("getParserResult: " + parserTime);
			LOGGER.info("startDocTime:    " + startDocumentTime);
			LOGGER.info("endDocTime:      " + endDocumentTime);
			LOGGER.info("startElement:    " + startElementTime);
			LOGGER.info("endElement:      " + endElementTime);
			LOGGER.info("charTime:        " + charTime);
			
			
			allResults.put("person", Arrays.asList(personResult));
			allResults.put("group", Arrays.asList(groupResult));
			allResults.put("organization", Arrays.asList(organizationResult));
			allResults.put("relation", relationResult);
			
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (SAXException e) {
			e.printStackTrace();
		}
		
		return allResults;
	}

}
