package no.ntnu.ime.libxsltgenerator;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.StringReader;
import java.io.StringWriter;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.nio.charset.Charset;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.logging.Logger;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import no.ntnu.ime.libxsltgenerator.exceptions.ValidSourceException;
import no.ntnu.ime.libxsltgenerator.generator.XSLTFactory;
import no.ntnu.ime.libxsltgenerator.models.SaxParserResult;
import no.ntnu.ime.libxsltgenerator.models.SaxParserResultData;
import no.ntnu.ime.libxsltgenerator.stats.PerformanceTimer;

import org.w3c.dom.Document;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

public class XSLTGenerator {

	private static final Logger LOGGER = Logger.getLogger(XSLTGenerator.class.getName());
	private StringBuilder performanceLog = new StringBuilder();

	public boolean isValidSourceURI(String sourceAdress)
			throws ValidSourceException {

		PerformanceTimer timer = PerformanceTimer.start();

		URL sourceURL;

		if (sourceAdress.isEmpty()) {
			return false;
		}

		try {
			sourceURL = new URL(sourceAdress);
		} catch (MalformedURLException e) {
			throw new ValidSourceException("Not a valid address");
		}

		try {
			sourceURL.openConnection().connect();
		} catch (IOException e) {
			throw new ValidSourceException(
					"Could not connect to the given address");
		}

		performanceLog.append("isValidSource took: " + timer.stop().getMsTime()
				+ "ms");

		return true;
	}

	public Map<String, List<SaxParserResult>> parseXMLSource(String url)
			throws Exception {
		PerformanceTimer timer = PerformanceTimer.start();

		if (url == null || url.isEmpty() || !isValidSourceURI(url)) {
			throw new RuntimeException("Invalid identifier");
		}

		File xmlFile = retrieveXMLSource(url);
		if (!xmlFile.exists() || !xmlFile.canRead()) {
			throw new RuntimeException("Unable to read file");
		}

		SAXHandler sh = new SAXHandler();

		Map<String, List<SaxParserResult>> result = sh.processSource(xmlFile);

		performanceLog.append("parseXMLSource took: "
				+ timer.stop().getMsTime() + "ms");

		return result;

	}

	public File retrieveXMLSource(String sourceAddress) throws RuntimeException {
		PerformanceTimer timer = PerformanceTimer.start();
		try {
			URL sourceURL = new URL(sourceAddress);
			String urlIdentifier = UUID.randomUUID().toString();

			URLConnection connection = sourceURL.openConnection();
			BufferedReader in = new BufferedReader(new InputStreamReader(
					connection.getInputStream(), Charset.forName("latin1")));

			StringBuilder stringBuilder = new StringBuilder();

			String line;
			while ((line = in.readLine()) != null) {
				stringBuilder.append(line);
			}
			in.close();

			String sourceFilename = getFilepathForUrlIdentifier(urlIdentifier);
			File tmpFile = new File(sourceFilename);

			FileOutputStream fos = new FileOutputStream(tmpFile);
			fos.write(stringBuilder.toString().getBytes());
			fos.close();

			LOGGER.info("retrieveXMLSource took: "
					+ timer.stop().getMsTime() + "ms");

			return tmpFile;

		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	private String getFilepathForUrlIdentifier(String urlIdentifier) {
		String tmpDir = System.getProperty("java.io.tmpdir");
		// Ensure we have trailing slash on both Unix and Windows systems
		if (!(tmpDir.endsWith("/") || tmpDir.endsWith("\\"))) {
			tmpDir += System.getProperty("file.separator");
		}
		String sourceFilename = tmpDir + urlIdentifier + ".xml";
		return sourceFilename;
	}

	public String getXlstForDataset(String xpath,
			Map<String, List<SaxParserResult>> xmlSource) {

		PerformanceTimer timer = PerformanceTimer.start();

		XSLTFactory generator = new XSLTFactory();

		if (xmlSource.containsKey("relation")) {
			for (SaxParserResult relationResult : xmlSource.get("relation")) {
				// Quick hack - TODO: Support multiple matches
				String sub = "";
				String obj = "";
				for (SaxParserResultData resData : relationResult.getEntries()) {
					if (sub.isEmpty()) {
						sub = resData.getQuery();
					} else {
						obj = resData.getQuery();
					}
				}

				String subjectPath = relationResult.getXpath() + "/" + sub;
				String objectPath = relationResult.getXpath() + "/" + obj;
				if (!subjectPath.equals(objectPath) && !objectPath.contains(subjectPath) && !subjectPath.contains(objectPath)){
					generator.addMembership(subjectPath, objectPath);
				}
			}
		}

		String personXpath = xmlSource.get("person").get(0).getXPath();
		generator.addPerson(personXpath, xmlSource.get("person").get(0)
				.getEntries());

		String xlst = generator.getTransformation().toString();

		try {
			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
			// Don't connect to the web for externals DTDs and such - we only
			// want to format, not validate.
			dbf.setFeature(
					"http://apache.org/xml/features/nonvalidating/load-dtd-grammar",
					false);
			dbf.setFeature(
					"http://apache.org/xml/features/nonvalidating/load-external-dtd",
					false);
			dbf.setValidating(false);
			DocumentBuilder db = dbf.newDocumentBuilder();
			Document document = db
					.parse(new InputSource(new StringReader(xlst)));
			Transformer transformer = TransformerFactory.newInstance()
					.newTransformer();
			transformer.setOutputProperty(OutputKeys.INDENT, "yes");
			transformer.setOutputProperty(OutputKeys.METHOD, "xml");
			transformer.setOutputProperty(
					"{http://xml.apache.org/xslt}indent-amount", "2");
			// initialize StreamResult with File object to save to file
			StreamResult result = new StreamResult(new StringWriter());
			DOMSource source = new DOMSource(document);
			transformer.transform(source, result);
			String xmlString = result.getWriter().toString();
			return xmlString;
		} catch (Exception e) {
			e.printStackTrace();
			return xlst;
		} finally {

			LOGGER.info("getXlstForDataset took: "
					+ timer.stop().getMsTime() + "ms");
		}

	}

	public String getXMLDocumentContentsForId(String uploadID) {
		BufferedReader in = null;
		StringBuilder stringBuilder = new StringBuilder();
		try {
			in = new BufferedReader(new InputStreamReader(new FileInputStream(
					new File(getFilepathForUrlIdentifier(uploadID)))));
			String line;
			while ((line = in.readLine()) != null) {
				stringBuilder.append(line);
			}
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			if (in != null) {
				try {
					in.close();
				} catch (IOException e) {
					// Nothing more we can do..
					e.printStackTrace();
				}
			}
		}

		return stringBuilder.toString();
	}

	public String attemptConversion(String result, File sourceXMLFile) {

		DOMSource xslt;
		DOMSource xmls;
		try {
			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance(); // .newDocumentBuilder();
			dbf.setNamespaceAware(true);

			xslt = new DOMSource(dbf.newDocumentBuilder().parse(
					new InputSource(new StringReader(result))));

			StringBuffer sb = new StringBuffer();
			BufferedReader br = new BufferedReader(
					new FileReader(sourceXMLFile));
			String line;
			while ((line = br.readLine()) != null) {
				sb.append(line + "\n");
			}
			String sourceXML = sb.toString();
			xmls = new DOMSource(dbf.newDocumentBuilder().parse(
					new InputSource(new StringReader(sourceXML))));

			Source xlstSource = xslt;
			Source xmlSource = xmls;

			StringWriter sw = new StringWriter();
			Result resultSource = new StreamResult(sw);

			TransformerFactory transFactory = TransformerFactory.newInstance();

			Transformer trans = transFactory.newTransformer(xlstSource);
			trans.setOutputProperty(OutputKeys.INDENT, "yes");
			trans.setOutputProperty(OutputKeys.METHOD, "xml");
			trans.setOutputProperty(
					"{http://xml.apache.org/xslt}indent-amount", "2");

			trans.transform(xmlSource, resultSource);
			return sw.toString();

		} catch (TransformerConfigurationException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (TransformerException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (SAXException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		} catch (ParserConfigurationException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		return "";
	}

	public String getStatistics() {
		return performanceLog.toString();
	}
}
